
//		ScreenShot.cpp - plugin to create screen shots


#define UNICODE					1
#define _UNICODE				1

#include <stdio.h>
#include <tchar.h>

#include <windows.h>
#include <GdiPlus.h>

#include "PlugIn.h"
#include "TakeCmd.h"


using namespace Gdiplus;


#define PLUGIN_NAME				L"ScreenShot"

#define VER_MAJOR				1
#define VER_MINOR				1
#define VER_BUILD				3
#define VER_PATCH				2

#ifdef _WIN64
#define PLATFORM				L"x64"
#else
#define PLATFORM				L"x86"
#endif

#define MAX_ARGS				1024
#define MAX_ARG_LEN				8192
#define SHORT_BUF_LEN			1024
#define MAX_FILENAME_LEN		32768

#define EXT_BUF_LEN				8

#define ERRHANDLE				GetStdHandle( STD_ERROR_HANDLE )

#define MAX_DISPLAY_MONITORS	64

#define FONT_SIZE_MIN			6
#define FONT_SIZE_MAX			96
#define FONT_SIZE_DEFAULT		12

#define NUM_HWA_WINDOW_CLASSES	3


#include "w3ccolors.cpp"



ULONG_PTR GdiplusToken;
GdiplusStartupInput StartupInput;



wchar_t LastFilename[MAX_FILENAME_LEN]		= L"";



#ifdef			SSWINLIST

wchar_t PluginFeatures[] = L"screenshot,sswinlist,_lastssfile";

#else		//	SSWINLIST not defined

wchar_t PluginFeatures[] = L"screenshot,_lastssfile";

#endif		//	SSWINLIST



const wchar_t ScreenShotHelpText[] = L"Save a picture of a window or the desktop.\n\
\n\
SCREENSHOT /A /D /F /H: /K: /M: /O /P: /Q /T /W: filename \r\n\
\n\
\t/A \tdesktop, all monitors\n\
\t/D \tdesktop\n\
\t/F \tforeground window\n\
\t/H:\twindow handle\n\
\t/K:\twindow class\n\
\t/M:\tmonitor number; implies /D \n\
\t/O \toverwrite\n\
\t/P:\tprocess ID\n\
\t/Q \tquietly\n\
\t/T \tTake Command\u2019s window\n\
\t/W:\twindow title\n\
\n\
Window title and class may use wildcards.  Process ID and window handle\n\
may be decimal, or hex with a leading 0X.  /K: /P: and /W: may be combined. \n\
Filename may be CLIP:  Default mode is /D.\n";


#ifdef			SSWINLIST

const wchar_t SSWinListHelpText[] = L"List visible top-level windows.\n\
\n\
SSWINLIST /K:class /P:pid /Q title \r\n\n\
\t/K:\twindow class\n\
\t/P:\tprocess ID\n\
\t/Q  \tno header and footer\n\
\n\
Window title and class may use wildcards.  Pid may be hex with a  \n\
leading 0X.  All args are optional.\n";

#endif		//	SSWINLIST


const wchar_t _LastSSFileHelpText[] = L"_LASTSSFILE \r\n\n\
Returns the filename of the last screen shot.\n";



#ifdef			SSWINLIST

const int NumHelpTexts							= 3;
const wchar_t HelpFeature[NumHelpTexts][12]		= { L"screenshot", L"sswinlist", L"_lastssfile" };
const wchar_t *HelpText[NumHelpTexts]			= { ScreenShotHelpText, SSWinListHelpText, _LastSSFileHelpText };

#else		//	SSWINLIST not defined

const int NumHelpTexts							= 2;
const wchar_t HelpFeature[NumHelpTexts][12]		= { L"screenshot", L"_lastssfile" };
const wchar_t *HelpText[NumHelpTexts]			= { ScreenShotHelpText, _LastSSFileHelpText };

#endif		//	SSWINLIST


const wchar_t ClipName[]				= L"CLIP:";
const wchar_t PrintName[]				= L"PRINT:";
const wchar_t ErrorString[]				= L"*Error*";
const wchar_t NullString[]				= L"(null)";


const wchar_t HWAWindowClass[NUM_HWA_WINDOW_CLASSES][30]	= { L"CASCADIA_HOSTING_WINDOW_CLASS", L"Chrome_WidgetWin_1", L"MozillaWindowClass" };



//	--------------------------------------------------------------------
//	Constants for graphics file types, file extensions, and mime types:

#define NUM_GRAF_EXTENSIONS		7
#define NUM_GRAF_TYPES			5
#define GRAF_TYPE_BMP			1
#define GRAF_TYPE_JPEG			2
#define GRAF_TYPE_PNG			3
#define GRAF_TYPE_GIF			4
#define GRAF_TYPE_TIFF			5

const wchar_t GrafExtension[NUM_GRAF_EXTENSIONS][6]		= { L".bmp", L".jpg", L".jpeg", L".png", L".gif", L".tif", L".tiff" };
unsigned int GrafType[NUM_GRAF_EXTENSIONS]				= { GRAF_TYPE_BMP, GRAF_TYPE_JPEG, GRAF_TYPE_JPEG, GRAF_TYPE_PNG, GRAF_TYPE_GIF, GRAF_TYPE_TIFF, GRAF_TYPE_TIFF };
const wchar_t GrafMimeType[NUM_GRAF_TYPES][11]			= { L"image/bmp", L"image/jpeg", L"image/png", L"image/gif", L"image/tiff" };

//	--------------------------------------------------------------------


#define MODE_NOT_SET			0
#define MODE_WINDOW				1
#define MODE_FG_WINDOW			2
#define MODE_HWND				3
#define MODE_DESKTOP			4

#define ALL_MONITORS			128

#define MIN_NEWSIZE				64
#define MAX_NEWSIZE				4096
#define MIN_PERCENTAGE			5
#define MAX_PERCENTAGE			200

#define MIN_COLUMNS				40
#define MAX_COLUMNS				512
#define DEFAULT_COLUMNS			100

#define WIN32_ERROR_CODE		0x8000




struct _ResizeParms {					//	Info for resizing output files:
	unsigned int NewWidth;				//	If NewWidth == 0 and NewHeight == 0, no resizing is done
	unsigned int NewHeight;				//	If NewHeight == 0 and Percentage is set, NewHeight defaults to NewWidth
	BOOL Percent;						//	if set, NewWidth and NewHeight are percentages and must be 5 .. 200
};


struct _DrawTextArgs {					//	Info for adding text to a bitmap:
	_DrawTextArgs *Next;				//	this might be an item in a linked list
	LPTSTR FaceName;					//	the desired typeface
	DWORD Color;						//	the text color
	unsigned int Flags;					//	bitmapped options
	unsigned int TextSize;				//	the font size
	unsigned int TextStyle;				//	0 regular, 1 bold, 2 italic, 4 underline
	int X;								//	position on the bitmap:
	int Y;								//	    if negative, reckon backward from the left/bottom
	wchar_t Text;						//	text to write begins here
};



int DisplayErrorHeader()
{
	wchar_t shortbuf[MAX_ARG_LEN] = L"%_batch";

	Qprintf( ERRHANDLE, L"%s plugin: ", PLUGIN_NAME );

	ExpandVariables( shortbuf, 1 );
	if ( shortbuf[0] == '0' )
		return 0;

	wcscpy_s( shortbuf, MAX_ARG_LEN, L"%_batchname [%_batchline]  " );
	ExpandVariables( shortbuf, 1 );
	Qprintf( ERRHANDLE, shortbuf );

	return 0;
}



#include "ParseArgs.cpp"



BOOL APIENTRY DllMain( HANDLE hModule, DWORD  dwReason, LPVOID lpReserved )
{
	switch( dwReason )	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
}



DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	static PLUGININFO piInfo;

	piInfo.pszDll			= (LPTSTR) PLUGIN_NAME;
	piInfo.pszAuthor		= (LPTSTR) L"Charles Dye";
	piInfo.pszEmail			= (LPTSTR) L"cdye@unm.edu";
	piInfo.pszWWW			= (LPTSTR) L"http://charlesdye.net/plugins/screenshot.html";
	piInfo.pszDescription	= (LPTSTR) L"ScreenShot plugin";
	piInfo.pszFunctions		= (LPTSTR) PluginFeatures;
	piInfo.nMajor			= VER_MAJOR;
	piInfo.nMinor			= VER_MINOR;
	piInfo.nBuild			= VER_BUILD;

	return &piInfo;
}



DLLExports BOOL WINAPI InitializePlugin( void )
{
	int rv = 0;

	rv = GdiplusStartup( &GdiplusToken, &StartupInput, NULL );
	if ( rv ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error %d initializing GDI+ -- plugin not loaded!\r\n", rv );
		return -1;
	}


	wchar_t shortbuf[SHORT_BUF_LEN] = L"%[_transient]%[_pipe]%[noloadmsg]";
	ExpandVariables( shortbuf, 1 );

	if ( wcscmp( shortbuf, L"00" ) == 0 )
		Printf( L"%s v%u.%u.%u loaded.\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );

	return 0;
}



DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{

#ifdef NEWHELP_VERSION
	CloseHelpWindow();
#endif

	GdiplusShutdown( GdiplusToken );
	return 0;
}




int ParseInt( LPTSTR inString, unsigned int *outValue, wchar_t **EndChar )
{
	if ( !inString || !outValue )
		return -1;

	unsigned int value = 0;
	wchar_t *sep1 = NULL;

	_set_errno( 0 );

	if ( ( inString[0] == '0' ) && ( towupper( inString[1] ) == 'X' ) && ( iswxdigit( inString[2] ) ) ) {
		//  parse hexadecimal

		value = wcstoul( inString + 2, &sep1, 16 );

		if ( errno )
			return -1;
		else if ( *sep1 && !EndChar )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}
	else if ( iswdigit( inString[0] ) ) {
		//  parse unsigned

		value = wcstoul( inString, &sep1, 10 );

		if ( errno )
			return -1;
		else if ( *sep1 && !EndChar )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}

	return -1;
}



int ParseSignedInt( LPTSTR inString, int *outValue, wchar_t **EndChar )
{
	if ( !inString || !outValue )
		return -1;

	int value = 0;
	wchar_t *sep1 = NULL;

	_set_errno( 0 );

	if ( ( inString[0] == '0' ) && ( towupper( inString[1] ) == 'X' ) && ( iswxdigit( inString[2] ) ) ) {
		//  parse hexadecimal

		value = wcstoul( inString + 2, &sep1, 16 );

		if ( errno )
			return -1;
		else if ( *sep1 && !EndChar )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}
	else if ( iswdigit( inString[0] ) || ( ( inString[0] == '-' ) && iswdigit( inString[1] ) ) ) {
		//  parse signed integer

		value = wcstol( inString, &sep1, 10 );

		if ( errno )
			return -1;
		else if ( *sep1 && !EndChar )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}

	return -1;
}



BOOL IsColonOrEquals( wchar_t ch )
{
	return ( ( ch == ':' ) || ( ch == '=' ) );
}



BOOL IsListSeparator( wchar_t ch )
{
	return ( ( ch == ',' ) || ( ch == ';' ) );
}



#include "FindMyWindow.cpp"



BOOL IsPrinter( LPTSTR String )
{
	if ( !String )
		return FALSE;

	if ( !_wcsnicmp( String, PrintName, wcslen( PrintName ) ) )
		return TRUE;
	else
		return FALSE;
}



int AppendFilename( LPTSTR Path, LPTSTR Filename, size_t BufferSize )
{
	if ( !Path || !Filename )
		return -1;

	
	BOOL NeedSlash		= FALSE;

	for ( size_t i = 0; Path[i]; i++ ) {
		if ( ( Path[i] == '/' ) || ( Path[i] == '\\' ) || ( Path[i] == ':' ) )
			NeedSlash = FALSE;
		else
			NeedSlash = TRUE;
	}

	if ( ( Filename[0] == '/' ) || ( Filename[0] == '\\' ) )
		NeedSlash = FALSE;
	
	if ( NeedSlash )
		wcscat_s( Path, BufferSize, L"\\" );

	wcscat_s( Path, BufferSize, Filename );
	return 0;
}



LPTSTR _ExtensionPart( LPTSTR Filename, LPTSTR outExt )
{
	if ( !Filename )
		return NULL;
	else if ( Filename[0] == '\0' )
		return NULL;


	LPTSTR Ext			= NULL;


	for ( size_t i = 0; Filename[i]; i++ ) {
		if ( ( Filename[i] == '/' ) || ( Filename[i] == '\\' ) || ( Filename[i] == ':' ) )
			Ext = NULL;
		else if ( Filename[i] == '.' )
			Ext = Filename + i;
	}

	if ( !Ext )
		return NULL;

	if ( outExt ) {
		wcscpy_s( outExt, MAX_FILENAME_LEN, Ext );
		return outExt;
	}

	return Ext;
}



//	Strips off any leading or trailing spaces from a string.

int TrimLine( LPTSTR String )
{
	if ( !String )
		return -1;
	else if ( String[0] == '\0' )
		return 0;

	int FirstNonSpace = -1, LastNonSpace = -1;
	wchar_t ch = '\0';

	for ( int i = 0; String[i]; i++ ) {
		ch = String[i];

		if ( !iswspace( ch ) ) {
			LastNonSpace = i;

			if ( FirstNonSpace == -1 )
				FirstNonSpace = i;
		}
	}

	if ( FirstNonSpace == -1 ) {
		String[0] = '\0';
		return 0;
	}

	String[LastNonSpace + 1] = '\0';

	if ( FirstNonSpace ) {
		int i = 0;

		do {
			ch = String[FirstNonSpace + i];
			String[i++] = ch;

		} while ( ch );
	}

	return 0;
}



int GetErrorText( int err, LPTSTR outString )
{
	if ( outString == NULL )
		return -1;

	if ( FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_MAX_WIDTH_MASK, NULL, err, 0, outString, SHORT_BUF_LEN, NULL ) == 0 )
		wsprintf( outString, L"Error %d", err );

	size_t l = wcslen( outString );

	if ( l > 0 )
		while ( iswspace( outString[l - 1] ) ) {
			l--;
			outString[l] = '\0';
		}

	return 0;
}



int GetGdiplusErrorText( int err, LPTSTR outString )
{
	if ( !outString )
		return -1;


	if ( ( err > 0 ) && ( err & WIN32_ERROR_CODE ) )
		return GetErrorText( err & 0x7fff, outString );


	int lasterr = GetLastError();
	if ( !lasterr )
		lasterr = -1;


	switch ( err ) {

	case -1 :
		wcscpy_s( outString, SHORT_BUF_LEN, L"Bad parameter (null pointer?)" );
		break;

	case Ok :
		wcscpy_s( outString, SHORT_BUF_LEN, L"Ok." );
		break;

	case GenericError :
		wcscpy_s( outString, SHORT_BUF_LEN, L"Generic error" );
		break;

	case InvalidParameter :
		wcscpy_s( outString, SHORT_BUF_LEN, L"Invalid parameter" );
		break;

	case OutOfMemory :
		wcscpy_s( outString, SHORT_BUF_LEN, L"Out of memory" );
		break;

	case ObjectBusy :
		wcscpy_s( outString, SHORT_BUF_LEN, L"Object is busy" );
		break;

	case InsufficientBuffer :
		wcscpy_s( outString, SHORT_BUF_LEN, L"Insufficient buffer" );
		break;

	case NotImplemented :
		wcscpy_s( outString, SHORT_BUF_LEN, L"Not implemented" );
		break;

	case Win32Error :
		GetErrorText( lasterr, outString );
		break;

	case WrongState :
		wcscpy_s( outString, SHORT_BUF_LEN, L"Wrong state" );
		break;

	case Aborted :
		wcscpy_s( outString, SHORT_BUF_LEN, L"Aborted" );
		break;

	case FileNotFound :
		wcscpy_s( outString, SHORT_BUF_LEN, L"File not found" );
		break;

	case ValueOverflow :
		wcscpy_s( outString, SHORT_BUF_LEN, L"Value overflow" );
		break;

	case AccessDenied :
		wcscpy_s( outString, SHORT_BUF_LEN, L"Access denied" );
		break;

	case UnknownImageFormat :
		wcscpy_s( outString, SHORT_BUF_LEN, L"Unknown image format" );
		break;

	case FontFamilyNotFound :
		wcscpy_s( outString, SHORT_BUF_LEN, L"Font family not found" );
		break;

	case FontStyleNotFound :
		wcscpy_s( outString, SHORT_BUF_LEN, L"Font style not found" );
		break;

	case NotTrueTypeFont :
		wcscpy_s( outString, SHORT_BUF_LEN, L"Not TrueType font" );
		break;

	case UnsupportedGdiplusVersion :
		wcscpy_s( outString, SHORT_BUF_LEN, L"Unsupported GDI+ version" );
		break;

	case GdiplusNotInitialized :
		wcscpy_s( outString, SHORT_BUF_LEN, L"GDI+ not initialized" );
		break;

	case PropertyNotFound :
		wcscpy_s( outString, SHORT_BUF_LEN, L"Property not found" );
		break;

	case PropertyNotSupported :
		wcscpy_s( outString, SHORT_BUF_LEN, L"Property not supported" );
		break;

	case 21 :
		wcscpy_s( outString, SHORT_BUF_LEN, L"Profile not found" );
		break;

	default :
		wsprintf( outString, L"GDI+ error %d", err );
	}

	return 0;
}



BOOL IsHWAWindow( HWND hWnd )
{
	if ( !hWnd )
		return FALSE;


	wchar_t ClassName[SHORT_BUF_LEN]		= L"";


	RealGetWindowClass( hWnd, ClassName, SHORT_BUF_LEN );
	if ( ClassName[0] == '\0' )
		return FALSE;

	for ( unsigned int i = 0; i < NUM_HWA_WINDOW_CLASSES; i++ )
		if ( !wcscmp( ClassName, HWAWindowClass[i] ) )
			return TRUE;

	return FALSE;
}



int ConsoleHandle( HANDLE inHandle )
{
	if ( !inHandle )
		return 0;

	if ( ( inHandle == (HANDLE) STD_INPUT_HANDLE ) || ( inHandle == (HANDLE) STD_OUTPUT_HANDLE ) || ( inHandle == (HANDLE) STD_ERROR_HANDLE ) )
		inHandle = GetStdHandle( (DWORD) inHandle );

	DWORD Mode = UINT_MAX;
	int rv = GetConsoleMode( inHandle, &Mode );

	if ( ( rv == 0 ) || ( Mode == UINT_MAX ) )
		return 0;
	else
		return 1;
}



int ConsoleHandle( DWORD inHandle )
{
	return ConsoleHandle( (HANDLE) inHandle );
}



int ConsoleWidth( )
{
	HANDLE ConsoleHandle = CreateFile( L"CONOUT$", ( GENERIC_READ | GENERIC_WRITE ), ( FILE_SHARE_READ | FILE_SHARE_WRITE ), 0, OPEN_EXISTING, NULL, NULL );
	CONSOLE_SCREEN_BUFFER_INFO ConsoleInfo;
	int Width = 0, rv = 0;

	if ( ConsoleHandle != INVALID_HANDLE_VALUE ) {
		rv = GetConsoleScreenBufferInfo( ConsoleHandle, &ConsoleInfo );
		CloseHandle( ConsoleHandle );

		if ( rv )
			Width = ConsoleInfo.dwSize.X;
	}

	if ( ( Width < MIN_COLUMNS ) || ( Width > MAX_COLUMNS ) )
		Width = 0;

	return Width;
}



int ResizeBitmap( Bitmap **Pic, _ResizeParms *Resize )
{
	if ( !Pic || !Resize )
		return -1;

	if ( !(*Pic ) )
		return -1;

	if ( ( Resize->NewWidth == 0 ) && ( Resize->NewHeight == 0 ) )
		return 0;


	unsigned int NewWidth		= Resize->NewWidth;
	unsigned int NewHeight		= Resize->NewHeight;

	unsigned int PicWidth		= (*Pic)->GetWidth();
	unsigned int PicHeight		= (*Pic)->GetHeight();
	if ( !PicWidth || !PicHeight )
		return ERROR_NO_DATA;

	PixelFormat SourcePxFormat	= (*Pic)->GetPixelFormat();
	float SourceHRes			= (*Pic)->GetHorizontalResolution();
	float SourceVRes			= (*Pic)->GetVerticalResolution();

	Bitmap *ResizeBmp			= NULL;
	Graphics *ResizeGraf		= NULL;

	int err						= 0;
	int rv						= 2;



	if ( Resize->Percent ) {

		if ( NewHeight == 0 )
			NewHeight = NewWidth;

		NewWidth	= ( PicWidth * NewWidth )   / 100;
		NewHeight	= ( PicHeight * NewHeight ) / 100;
	}
	else {
		double ratio = (double) PicWidth / (double) PicHeight;

		if ( NewHeight == 0 )
			NewHeight = (unsigned int) ( (double) NewWidth / ratio );
		else if ( NewWidth == 0 )
			NewWidth = (unsigned int) ( (double) NewHeight * ratio );
	}

	if ( ( NewWidth == PicWidth ) && ( NewHeight == PicHeight ) )
		return 0;


	ResizeBmp = new Bitmap( NewWidth, NewHeight, SourcePxFormat );
	if ( ResizeBmp ) {

		ResizeBmp->SetResolution( SourceHRes, SourceVRes );

		ResizeGraf = new Graphics( ResizeBmp );

		if ( ResizeGraf ) {

			ResizeGraf->SetInterpolationMode( InterpolationModeHighQuality );

			err = ResizeGraf->DrawImage( (*Pic), 0, 0, NewWidth, NewHeight );

			if ( !err ) {
				delete (*Pic);
				(*Pic) = ResizeBmp;
				rv = 0;
			}

			delete ResizeGraf;
		}
	
		if ( rv )
			delete ResizeBmp;
	}

	return rv;
}



int DrawTextOnBitmap( Bitmap *Pic, _DrawTextArgs *Args )
{
	if ( !Pic || !Args )
		return -1;


	wchar_t DefaultFace[]	= L"Verdana";
	LPTSTR FaceName			= DefaultFace;

	Graphics *TextGraf		= NULL;
	Font *TextFont			= NULL;
	SolidBrush TextBrush( Color( (BYTE) ( ( Args->Color & 0xff0000 ) >> 16 ), (BYTE) ( ( Args->Color & 0x00ff00 ) >> 8 ), (BYTE) ( Args->Color & 0x0000ff ) ) );
	PointF Origin( 0.0f, 0.0f );

	int rv					= 0;



	if ( Args->FaceName )
		if ( Args->FaceName[0] )
			FaceName = Args->FaceName;


	if ( Args->X >= 0 )
		Origin.X = (float) ( Args->X );
	else
		Origin.X = (float) ( Pic->GetWidth() + Args->X );

	if ( Args->Y >= 0 )
		Origin.Y = (float) ( Args->Y );
	else
		Origin.Y = (float) ( Pic->GetHeight() + Args->Y );



	TextFont = new Font( FaceName, (float) Args->TextSize, Args->TextStyle, UnitPixel, NULL );
	if ( !TextFont )
		return rv;


	TextGraf = new Graphics( Pic );

	if ( TextGraf ) {
		if ( TextGraf->DrawString( &( Args->Text ), -1, TextFont, Origin, &TextBrush ) == Ok )
			rv = 0;

		delete( TextGraf );
	}

	delete( TextFont );

	return rv;
}



int DisplayWindowNotFoundError( DWORD Pid, LPTSTR Title, LPTSTR Class )
{
	BOOL AnySpecs		= FALSE;


	DisplayErrorHeader();

	Qprintf( ERRHANDLE, L"Can\u2019t find window " );

	if ( Title )
		if ( Title[0] ) {
			Qprintf( ERRHANDLE, L"\"%s\" ", Title );
			AnySpecs = TRUE;
		}

	if ( Class )
		if ( Class[0] ) {
			Qprintf( ERRHANDLE, L"\u00ab%s\u00bb ", Class );
			AnySpecs = TRUE;
		}

	if ( Pid ) {
		Qprintf( ERRHANDLE, L"for PID 0x%04X", Pid );
		AnySpecs = TRUE;
	}

	if ( !AnySpecs )
		Qprintf( ERRHANDLE, L"(unspecified)" );

	Qprintf( ERRHANDLE, L"\r\n" );
	return 0;
}



int IsGraphicsExtension( LPTSTR Ext )
{
	if ( !Ext )
		return 0;

	for ( unsigned int i = 0; i < NUM_GRAF_EXTENSIONS; i++ )
		if ( !_wcsicmp( Ext, GrafExtension[i] ) )
			return GrafType[i];

	return 0;
}



int GetDefaultExtension( LPTSTR outExt )
{
	wchar_t shortbuf[SHORT_BUF_LEN]		= L"";
	int rv = 0;


	GetEnvironmentVariable( L"ScreenShotExt", shortbuf, SHORT_BUF_LEN );
	
	rv = IsGraphicsExtension( shortbuf );

	if ( rv )
		wcscpy_s( outExt, EXT_BUF_LEN, shortbuf );
	else
		wcscpy_s( outExt, EXT_BUF_LEN, L".png" );

	return rv;
}



#include "ForceFGWin.cpp"



struct _GetMonitorCoordsStruct {
	unsigned int NumMonitors;
	unsigned int Default;
	int Left[MAX_DISPLAY_MONITORS];
	int Top[MAX_DISPLAY_MONITORS];
	int Width[MAX_DISPLAY_MONITORS];
	int Height[MAX_DISPLAY_MONITORS];
};



BOOL CALLBACK GetMonitorCoordsEnumProc( HMONITOR hMonitor, HDC hdcMonitor, LPRECT lprcMonitor, LPARAM Monitors )
{
	if ( !Monitors )
		return TRUE;

	_GetMonitorCoordsStruct *monitors = (_GetMonitorCoordsStruct*) Monitors;

	if ( monitors->NumMonitors >= MAX_DISPLAY_MONITORS )
		return FALSE;

	//	wprintf( L"   * Monitor %2u origin:  (%d,%d)\n", monitors->NumMonitors, lprcMonitor->left, lprcMonitor->top );

	monitors->Left[monitors->NumMonitors]		= lprcMonitor->left;
	monitors->Top[monitors->NumMonitors]		= lprcMonitor->top;
	monitors->NumMonitors++;

	return monitors->NumMonitors < MAX_DISPLAY_MONITORS;
}



int GetMonitorCoords( _GetMonitorCoordsStruct *Monitors )
{
	if ( !Monitors )
		return -1;


	unsigned int DevIndex		= 0;
	DISPLAY_DEVICE DispDevice;
	DEVMODE DevMode;


	memset( Monitors, 0x00, sizeof( _GetMonitorCoordsStruct ) );
	Monitors->NumMonitors	= 0;


	if ( !EnumDisplayMonitors( NULL, NULL, GetMonitorCoordsEnumProc, (LPARAM) Monitors ) )
		return 1;

	for ( unsigned int i = 0; ; i++ ) {
		memset( &DispDevice, 0x00, sizeof( DISPLAY_DEVICE ) );
		DispDevice.cb = sizeof( DISPLAY_DEVICE );

		if ( !EnumDisplayDevices( NULL, i, &DispDevice, 0 ) )
			break;

		if ( ( DispDevice.StateFlags & ( DISPLAY_DEVICE_ACTIVE | DISPLAY_DEVICE_MIRRORING_DRIVER ) ) == DISPLAY_DEVICE_ACTIVE ) {

			memset( &DevMode, 0x00, sizeof( DEVMODE ) );
			DevMode.dmSize = sizeof( DEVMODE );

			if ( EnumDisplaySettings( DispDevice.DeviceName, ENUM_CURRENT_SETTINGS, &DevMode ) ) {
				//	wprintf( L"   * Monitor %2u :  '%s'  %u x %u\n", DevIndex, DispDevice.DeviceName, DevMode.dmPelsWidth, DevMode.dmPelsHeight );
				Monitors->Width[DevIndex]	= DevMode.dmPelsWidth;
				Monitors->Height[DevIndex]	= DevMode.dmPelsHeight;
			}

			if ( DispDevice.StateFlags & DISPLAY_DEVICE_PRIMARY_DEVICE )
				Monitors->Default = DevIndex;

			DevIndex++;
		}
	}

	return 0;
}



struct _FindWindowStruct {
	DWORD Pid;
	LPTSTR Title;
	LPTSTR Class;
	HWND WindowHandle;
};



BOOL CALLBACK MyFindWindowProc( HWND hWnd, LPARAM lParam )
{
	_FindWindowStruct *FindStruct = (_FindWindowStruct*) lParam;


	wchar_t Title[MAX_ARG_LEN]		= L"";
	wchar_t Class[SHORT_BUF_LEN]	= L"";
	DWORD Pid						= 0;


	if ( !IsWindowVisible( hWnd ) )
		return TRUE;


	if ( FindStruct->Pid ) {
		GetWindowThreadProcessId( hWnd, &Pid );

		if ( Pid != FindStruct->Pid )
			return TRUE;
	}

	if ( FindStruct->Title ) {
		GetWindowText( hWnd, Title, MAX_ARG_LEN );

		if ( WildcardComparison( FindStruct->Title, Title, 0, 1 ) )
			return TRUE;
	}

	if ( FindStruct->Class ) {
		RealGetWindowClass( hWnd, Class, SHORT_BUF_LEN );

		if ( WildcardComparison( FindStruct->Class, Class, 0, 1 ) )
			return TRUE;
	}
	
	FindStruct->WindowHandle = hWnd;
	return FALSE;
}



HWND MyFindWindow( DWORD Pid, LPTSTR Title, LPTSTR Class )
{
	if ( !Pid && !Title && !Class )
		return NULL;


	_FindWindowStruct FindStruct;


	FindStruct.Pid			= Pid;
	FindStruct.Title		= NULL;
	FindStruct.Class		= NULL;
	FindStruct.WindowHandle = NULL;


	if ( Title )
		if ( Title[0] )
			FindStruct.Title = Title;

	if ( Class )
		if ( Class[0] )
			FindStruct.Class = Class;


	EnumWindows( MyFindWindowProc, (LPARAM) &FindStruct );

	if ( ( FindStruct.WindowHandle == NULL ) && ( Pid == GetCurrentProcessId() ) )
		return FindMyWindow();

	return FindStruct.WindowHandle;
}



int MassageLastError()
{
	int err		= GetLastError();


	if ( err )
		return err | WIN32_ERROR_CODE;
	else
		return -1;
}



int GdiplusBitmapToClipboard( Bitmap *Pic )
{
	if ( !Pic )
		return -1;


	HBITMAP TempPic			= NULL;
	HBITMAP SavePic			= NULL;
	
	int rv					= 0;


	rv = Pic->GetHBITMAP( 0x000000, &TempPic );

	if ( TempPic ) {
		SavePic = (HBITMAP) CopyImage( TempPic, IMAGE_BITMAP, 0, 0, LR_DEFAULTSIZE );

		if ( SavePic ) {
			if ( OpenClipboard( NULL ) ) {
				if ( EmptyClipboard() ) {

					if ( SetClipboardData( CF_BITMAP, SavePic ) ) {
						SavePic = NULL;
						rv = 0;
					}
					else
						rv = MassageLastError();		//	SetClipboardData() failed
				}
				else
					rv = MassageLastError();			//	EmptyClipboard() failed

				CloseClipboard();
			}
			else
				rv = MassageLastError();				//	OpenClipboard() failed

			if ( SavePic )
				DeleteObject( SavePic );
		}
		else
			rv = MassageLastError();					//	CopyImage() failed

		DeleteObject( TempPic );
	}
	else {												//	GetHBITMAP() failed:
		if ( !rv )
			rv = -1;
	}

	return rv;
}



int PrintImage( Image *PrintImage, HDC PrintDC )
{
	Image *SourceImage = NULL;

	SourceImage = PrintImage->Clone();

	if ( SourceImage == NULL )
		return -1;

	int rv = 0;


	int DPI_X = 96, DPI_Y = 96;

	int PrintPixelsX = 0, PrintPixelsY = 0;
	int PrintOriginX = 0, PrintOriginY = 0;
	int LogPixelsX = 0, LogPixelsY = 0;

	PrintPixelsX	= GetDeviceCaps( PrintDC, HORZRES );
	PrintPixelsY	= GetDeviceCaps( PrintDC, VERTRES );
	LogPixelsX		= GetDeviceCaps( PrintDC, LOGPIXELSX );
	LogPixelsY		= GetDeviceCaps( PrintDC, LOGPIXELSY );
	PrintOriginX	= GetDeviceCaps( PrintDC, PHYSICALOFFSETX );
	PrintOriginY	= GetDeviceCaps( PrintDC, PHYSICALOFFSETY );

	PrintOriginX = PrintOriginX * DPI_X / LogPixelsX;
	PrintOriginY = PrintOriginY * DPI_Y / LogPixelsY;

	int PrintGrafX = PrintPixelsX * DPI_X / LogPixelsX;
	int PrintGrafY = PrintPixelsY * DPI_Y / LogPixelsY;

	//  wprintf( L"   * PrintGraf = %d x %d (%d x %d for image)\r\n", PrintGrafX, PrintGrafY, PrintGrafX, PrintGrafY2 );
	//  wprintf( L"   * Print pixels:  %d x %d (per inch)\r\n", LogPixelsX, LogPixelsY );
	//  wprintf( L"   * Print origin:  %d x %d\r\n", PrintOriginX, PrintOriginY );

	int ImageX = SourceImage->GetWidth();
	int ImageY = SourceImage->GetHeight();
	int OriginX = 0, OriginY = 0;

	float ImageRatio = (float) ImageX / (float) ImageY;
	float GrafRatio  = (float) PrintGrafX / (float) PrintGrafY;
	BOOL TurnToFit = FALSE;


	if ( ( ImageRatio > 1.0 ) && ( GrafRatio < 1.0 ) )
		TurnToFit = TRUE;
	else if ( ( ImageRatio < 1.0 ) && ( GrafRatio > 1.0 ) )
		TurnToFit = TRUE;
	
	
	Graphics *PrintGraf = NULL;
	PrintGraf = new Graphics( PrintDC );
	if ( PrintGraf == NULL ) {
		delete SourceImage;
		return 2;
	}

	
	StartPage( PrintDC );

	//  draw the image:

	if ( TurnToFit && ( ( ImageX > PrintGrafX ) || ( ImageY > PrintGrafY ) ) ) {
		//  image is too big and image axis is opposite; turn image to fit:

		SourceImage->RotateFlip( Rotate90FlipNone );
		ImageX = SourceImage->GetWidth();
		ImageY = SourceImage->GetHeight();
		ImageRatio = (float) ImageX / (float) ImageY;
	}

	if ( ( ImageX > PrintGrafX ) || ( ImageY > PrintGrafY ) ) {
		//  image is still too big; compute a new size:

		float RatioX = (float) PrintGrafX / (float) ImageX;
		float RatioY = (float) PrintGrafY / (float) ImageY;

		if ( RatioX > RatioY )
			RatioX = RatioY;

		ImageX = (int) ( (float) ImageX * RatioX );
		ImageY = (int) ( (float) ImageY * RatioX );
	}

	if ( ImageX < PrintGrafX )
		OriginX = ( PrintGrafX - ImageX ) / 2;

	if ( ImageY < PrintGrafY )
		OriginY = ( PrintGrafY - ImageY ) / 2;

	//  wprintf( L"   * Image size:  %d x %d\r\n", ImageX, ImageY );
	//  wprintf( L"   * Print area:  %d x %d\r\n", PrintGrafX, PrintGrafY );
	//  wprintf( L"   * Origin:  %d x %d\r\n", OriginX, OriginY );

	PrintGraf->SetInterpolationMode( InterpolationModeHighQuality );
	PrintGraf->DrawImage( SourceImage, PrintOriginX + OriginX, PrintOriginY + OriginY, ImageX, ImageY );


	EndPage( PrintDC );
	delete PrintGraf;
	delete SourceImage;

	return 0;
}



int PrintGdiplusBitmap( Bitmap *Pic, LPTSTR Printer )
{
	if ( !Pic )
		return -1;


	wchar_t PrinterName[SHORT_BUF_LEN]		= L"";
	DWORD BufLen							= SHORT_BUF_LEN;

	HANDLE PrinterHandle					= NULL;
	DEVMODE *PrintDevMode					= NULL;
	LONG DevModeSize						= 0;
	HDC PrintDC								= NULL;
	DOCINFO DocInfo;
	wchar_t DocName[SHORT_BUF_LEN]			= L"";

	int rv = 0;



	if ( Printer )
		if ( Printer[0] )
			wcscpy_s( PrinterName, SHORT_BUF_LEN, Printer );

	if ( PrinterName[0] == '\0' )
		GetDefaultPrinter( PrinterName, &BufLen );

	if ( PrinterName[0] == '\0' )
		return ERROR_FILE_NOT_FOUND | WIN32_ERROR_CODE;
	

	wsprintf( DocName, L"%s %u.%u.%u", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );
	DocInfo.lpszDocName = DocName;


	if ( OpenPrinter( PrinterName, &PrinterHandle, NULL ) ) {

		DevModeSize = DocumentProperties( GetConsoleWindow(), PrinterHandle, PrinterName, NULL, NULL, 0 );

		if ( DevModeSize ) {

			PrintDevMode = (DEVMODE*) malloc( DevModeSize );

			if ( PrintDevMode ) {

				if ( DocumentProperties( GetConsoleWindow(), PrinterHandle, PrinterName, PrintDevMode, NULL, DM_OUT_BUFFER ) == IDOK ) {

					PrintDC = CreateDC( NULL, PrinterName, NULL, PrintDevMode );

					if ( PrintDC ) {

						StartDoc( PrintDC, &DocInfo );

						rv = PrintImage( Pic, PrintDC );

						EndDoc( PrintDC );
						DeleteDC( PrintDC );
					}
					else
						rv = MassageLastError();
				}
				else
					rv = MassageLastError();

				free( PrintDevMode );
			}
			else
				rv = ERROR_NOT_ENOUGH_MEMORY | WIN32_ERROR_CODE;
		}
		else
			rv = MassageLastError();

		ClosePrinter( PrinterHandle );
	}
	else
		rv = MassageLastError();


	return rv;
}



int GetEncoderClsid( const WCHAR* format, CLSID* pClsid )
{
	UINT  num = 0;          // number of image encoders
	UINT  size = 0;         // size of the image encoder array in bytes

	ImageCodecInfo* pImageCodecInfo = NULL;

	GetImageEncodersSize( &num, &size );
   
	if ( !size )
		return -1;  // Failure

	pImageCodecInfo = (ImageCodecInfo*) malloc( size );
	if ( !pImageCodecInfo )
		return -1;  // Failure

	GetImageEncoders( num, size, pImageCodecInfo );

	for ( UINT j = 0; j < num; ++j ) {
		if ( !wcscmp( pImageCodecInfo[j].MimeType, format ) ) {
			*pClsid = pImageCodecInfo[j].Clsid;
			free( pImageCodecInfo );
			return j;  // Success
		}    
	}

	free( pImageCodecInfo );
	return -1;  // Failure
}



int GetEncoderClsidByExtension( LPTSTR Filename, CLSID *pClsid )
{
	if ( !Filename || !pClsid )
		return 0;


	int rv		= IsGraphicsExtension( _ExtensionPart( Filename, NULL ) );

	if ( rv )
		GetEncoderClsid( GrafMimeType[rv - 1], pClsid );

	return rv;
}



int SaveWindowPic( HDC WindowDC, LPTSTR Filename, BOOL Overwrite, int Left, int Top, int Width, int Height, _ResizeParms *Resize, _DrawTextArgs *DrawText )
{
	if ( !WindowDC || !Filename )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	CLSID EncoderClsid;
	HDC CaptureDC				= NULL;
	int ColorDepth				= 0;
	HBITMAP CaptureBM			= NULL;
	HGDIOBJ OldObject			= NULL;

	Bitmap *Snapshot			= NULL;
	EncoderParameters EncoderParams;
	ULONG Quality				= 100;

	int rv = 1;



	if ( _wcsicmp( Filename, ClipName ) && !IsPrinter( Filename ) ) {

		if ( GetEncoderClsidByExtension( Filename, &EncoderClsid ) == 0 )
			return UnknownImageFormat;

		if ( QueryIsFile( Filename ) && !Overwrite )
			return ERROR_FILE_EXISTS | WIN32_ERROR_CODE;
	}



	EncoderParams.Count		= 1;
	EncoderParams.Parameter[0].NumberOfValues	= 1;
	EncoderParams.Parameter[0].Guid				= EncoderQuality;
	EncoderParams.Parameter[0].Type				= EncoderParameterValueTypeLong;
	EncoderParams.Parameter[0].Value			= &Quality;


	CaptureDC = CreateCompatibleDC( WindowDC );

	if ( CaptureDC ) {
		CaptureBM = CreateCompatibleBitmap( WindowDC, Width, Height );

		if ( CaptureBM ) {
			OldObject = SelectObject( CaptureDC, CaptureBM );

			if ( OldObject && ( OldObject != HGDI_ERROR ) ) {

				if ( BitBlt( CaptureDC, 0, 0, Width, Height, WindowDC, Left, Top, SRCCOPY ) ) {

					Snapshot = new Bitmap( CaptureBM, (HPALETTE) NULL );
					if ( Snapshot ) {

						if ( Resize )
							if ( ( Resize->NewWidth ) || ( Resize->NewHeight ) ) 
								ResizeBitmap( &Snapshot, Resize );

						while ( DrawText ) {
							DrawTextOnBitmap( Snapshot, DrawText );
							DrawText = DrawText->Next;
						}

						if ( _wcsicmp( Filename, ClipName ) == 0 )
							rv = GdiplusBitmapToClipboard( Snapshot );
						else if ( IsPrinter( Filename ) )
							rv = PrintGdiplusBitmap( Snapshot, Filename + wcslen( PrintName ) );
						else
							rv = Snapshot->Save( Filename, &EncoderClsid, &EncoderParams );

						delete Snapshot;
					}
				}

				SelectObject( CaptureDC, OldObject );
			}

			DeleteObject( CaptureBM );
		}

		DeleteDC( CaptureDC );
	}

	return rv;
}



int GenerateDefaultFilename( LPTSTR outFilename, LPTSTR Dir, LPTSTR ext )
{
	if ( !outFilename )
		return -1;


	wchar_t full[MAX_FILENAME_LEN]		= L"";
	wchar_t name[SHORT_BUF_LEN]			= L"";
	SYSTEMTIME LocalTime;


	GetLocalTime( &LocalTime );


	swprintf( name, SHORT_BUF_LEN, L"%s-%4u%02u%02u-%02u%02u%02u%s", PLUGIN_NAME, LocalTime.wYear, LocalTime.wMonth, LocalTime.wDay, LocalTime.wHour, LocalTime.wMinute, LocalTime.wSecond, ext ? ext : L"" );


	if ( Dir )
		if ( Dir[0] )
			wcscpy_s( full, MAX_FILENAME_LEN, Dir );

	if ( full[0] == '\0' )
		GetEnvironmentVariable( L"ScreenShotPath", full, MAX_FILENAME_LEN );

	if ( full[0] == '\0' )
		GetCurrentDirectory( MAX_FILENAME_LEN, full );

	AppendFilename( full, name, MAX_FILENAME_LEN );
	wcscpy_s( outFilename, MAX_FILENAME_LEN, full );
	return 0;
}



int ShowCmdHelp( LPCTSTR HelpText, BOOL ShowVersion )
{
	QPuts( (LPTSTR) HelpText );

	if ( ShowVersion )
		Printf( L"\r\n\t\t\t\t%s  %u.%u.%u.%u (%s)\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, VER_PATCH, PLATFORM );

	return 0;
}



DLLExports int WINAPI screenshot( LPTSTR inString )
{
	if ( !inString )
		return ERROR_INVALID_PARAMETER;


	int Mode								= MODE_DESKTOP;

	LPTSTR WindowTitle						= NULL;
	LPTSTR WindowClass						= NULL;
	wchar_t DefaultExtension[EXT_BUF_LEN]	= L"";
	wchar_t Filename[MAX_FILENAME_LEN]		= L"";
	wchar_t ErrorText[SHORT_BUF_LEN]		= L"";

	int TextX								= 0;
	int TextY								= -36;
	wchar_t TextFace[SHORT_BUF_LEN]			= L"Verdana";
	unsigned int TextSize					= 24;
	unsigned int TextStyle					= 0;
	DWORD TextColor							= 0xffffff;
	_DrawTextArgs *FirstTextArgs			= NULL;

	unsigned int Monitor					= ALL_MONITORS - 1;
	unsigned int Pid						= 0;

	BOOL ClientAreaOnly						= FALSE;
	BOOL WindowsOwnContext					= FALSE;
	BOOL Overwrite							= FALSE;
	BOOL Quiet								= FALSE;
	BOOL Debug								= FALSE;
	BOOL Minimized							= FALSE;

	HWND WindowHandle						= NULL;
	WINDOWINFO WindowInfo;

	HDC WindowDC							= NULL;

	int Left								= 0;
	int Top									= 0;
	int Width								= 0;
	int Height								= 0;

	HWND ForegroundWindow					= 0;

	int rv									= 1;


	_ResizeParms ResizeArgs = { 0, 0, FALSE };


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {

			if ( !wcsncmp( arg[i], L"/?", 2 ) ) {
				ShowCmdHelp( ScreenShotHelpText, ( arg[i][2] == '?' ) );
				rv = 0;
				goto cleanup_and_exit;
			}
			else if ( !wcscmp( arg[i], L"/$" ) )
				Debug = TRUE;
			else if ( !_wcsicmp( arg[i], L"/A" ) )
				Mode = MODE_DESKTOP, Monitor = ALL_MONITORS;
			else if ( !_wcsicmp( arg[i], L"/C" ) )
				ClientAreaOnly= TRUE;
			else if ( !_wcsicmp( arg[i], L"/D" ) )
				Mode = MODE_DESKTOP;
			else if ( !_wcsicmp( arg[i], L"/E" ) )
				WindowsOwnContext = TRUE;
			else if ( !_wcsicmp( arg[i], L"/F" ) || !_wcsicmp( arg[i], L"/FG" ) )
				Mode = MODE_FG_WINDOW;
			else if ( !_wcsnicmp( arg[i], L"/H", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, (unsigned int*) &WindowHandle, NULL ) ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Bad HWND value: \"%s\"\r\n", arg[i] );
					goto cleanup_and_exit;
				}

				if ( !WindowHandle ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"HWND cannot be zero\r\n", arg[i] );
					goto cleanup_and_exit;
				}

				Mode = MODE_HWND;
			}
			else if ( !_wcsnicmp( arg[i], L"/K", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] == '\0' ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Missing window class\r\n" );
					goto cleanup_and_exit;
				}
				
				WindowClass = arg[i] + j;
				Mode = MODE_WINDOW;
			}
			else if ( !_wcsnicmp( arg[i], L"/M", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &Monitor, NULL ) ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad monitor value: \"%s\"\r\n", arg[i] );
					goto cleanup_and_exit;
				}

				Mode = MODE_DESKTOP;
			}
			else if ( !_wcsicmp( arg[i], L"/O" ) )
				Overwrite = TRUE;
			else if ( !_wcsnicmp( arg[i], L"/P", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &Pid, NULL ) ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad PID value: \"%s\"\r\n", arg[i] );
					goto cleanup_and_exit;
				}

				if ( Pid == 0 ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"PID cannot be zero\r\n" );
					goto cleanup_and_exit;
				}

				Mode = MODE_WINDOW;
			}
			else if ( !_wcsicmp( arg[i], L"/Q" ) )
				Quiet = TRUE;
			else if ( !_wcsnicmp( arg[i], L"/R", 2 ) ) {
				_wcsupr_s( arg[i], MAX_ARG_LEN );

				wchar_t *sep1 = arg[i] + 2;
				if ( IsColonOrEquals( *sep1 ) )
					sep1++;


				if ( *sep1 ) {

					if ( IsListSeparator( *sep1 ) ) {
						ResizeArgs.NewWidth	= 0;
						ResizeArgs.Percent	= FALSE;
						sep1++;

						if ( ParseInt( sep1, &(ResizeArgs.NewHeight ), NULL ) ) {
							DisplayErrorHeader();
							Qprintf( ERRHANDLE, L"Bad NewHeight value: \"%s\"\r\n", arg[i] );
							goto cleanup_and_exit;
						}
					}
					else if ( ParseInt( sep1, &(ResizeArgs.NewWidth), &sep1 ) ) {
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Bad NewWidth value: \"%s\"\r\n", arg[i] );
						goto cleanup_and_exit;
					}

					if ( *sep1 == '%' ) {
						ResizeArgs.Percent = TRUE;

						sep1++;

						if ( *sep1 == '\0' )
							ResizeArgs.NewHeight = ResizeArgs.NewWidth;
						else if ( IsListSeparator( *sep1 ) ) {
							sep1++;

							if ( ParseInt( sep1, &(ResizeArgs.NewHeight), &sep1 ) ) {
								DisplayErrorHeader();
								Qprintf( ERRHANDLE, L"Bad NewHeight value: \"%s\"\r\n", arg[i] );
								goto cleanup_and_exit;
							}

							if ( wcscmp( sep1, L"%" ) ) {
								DisplayErrorHeader();
								Qprintf( ERRHANDLE, L"If NewWidth is a percentage, then NewHeight must be a percentage also\r\n" );
								goto cleanup_and_exit;
							}
						}
						else if ( *sep1 ) {
							DisplayErrorHeader();
							Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", sep1 );
							goto cleanup_and_exit;
						}

					}
					else if ( *sep1 == '\0' ) {
						ResizeArgs.NewHeight = 0;
						ResizeArgs.Percent = FALSE;
					}
					else if ( IsListSeparator( *sep1 ) ) {
						sep1++;

						if ( ParseInt( sep1, &(ResizeArgs.NewHeight), NULL ) ) {
							DisplayErrorHeader();
							Qprintf( ERRHANDLE, L"Bad NewHeight value: \"%s\"\r\n", arg[i] );
							goto cleanup_and_exit;
						}
					}


					if ( ResizeArgs.NewWidth || ResizeArgs.NewHeight ) {

						if ( ResizeArgs.Percent ) {				//	validage percentage resize args:

							if ( ( ResizeArgs.NewWidth < MIN_PERCENTAGE ) || ( ResizeArgs.NewWidth > MAX_PERCENTAGE ) ) {
								DisplayErrorHeader();
								Qprintf( ERRHANDLE, L"NewWidth must be %u%% - %u%% :  %u%%\r\n", MIN_PERCENTAGE, MAX_PERCENTAGE, ResizeArgs.NewWidth );
								goto cleanup_and_exit;
							}

							if ( ( ResizeArgs.NewHeight < MIN_PERCENTAGE ) || ( ResizeArgs.NewHeight > MAX_PERCENTAGE ) ) {
								DisplayErrorHeader();
								Qprintf( ERRHANDLE, L"NewHeight must be %u%% - %u%% :  %u%%\r\n", MIN_PERCENTAGE, MAX_PERCENTAGE, ResizeArgs.NewHeight );
								goto cleanup_and_exit;
							}
						}
						else {									//	validate absolute resize args:

							if ( ResizeArgs.NewWidth && ( ( ResizeArgs.NewWidth < MIN_NEWSIZE ) || ( ResizeArgs.NewWidth > MAX_NEWSIZE ) ) ) {
								DisplayErrorHeader();
								Qprintf( ERRHANDLE, L"NewWidth must be %u - %u:  %u\r\n", MIN_NEWSIZE, MAX_NEWSIZE, ResizeArgs.NewWidth );
								goto cleanup_and_exit;
							}

							if ( ResizeArgs.NewHeight && ( ( ResizeArgs.NewHeight < MIN_NEWSIZE ) || ( ResizeArgs.NewHeight > MAX_NEWSIZE ) ) ) {
								DisplayErrorHeader();
								Qprintf( ERRHANDLE, L"NewHeight must be %u - %u:  %u\r\n", MIN_NEWSIZE, MAX_NEWSIZE, ResizeArgs.NewHeight );
								goto cleanup_and_exit;
							}
						}
					}

				}
				else {
					ResizeArgs.NewWidth		= 0;
					ResizeArgs.NewHeight	= 0;
					ResizeArgs.Percent		= FALSE;
				}
			}
			else if ( !_wcsicmp( arg[i], L"/T" ) || !_wcsicmp( arg[i], L"/TCMD" ) ) {
				WindowHandle = FindMyWindow();
				Mode = MODE_HWND;

				if ( !WindowHandle ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Can\u2019t find Take Command\u2019s window\r\n" );
					rv = 2;
					goto cleanup_and_exit;
				}
			}
			else if ( !_wcsnicmp( arg[i], L"/W", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] == '\0' ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Missing window title\r\n" );
					goto cleanup_and_exit;
				}
				
				WindowTitle = arg[i] + j;
				Mode = MODE_WINDOW;
			}
			else if ( !_wcsnicmp( arg[i], L"/XC", 3 ) && IsColonOrEquals( arg[i][3] ) ) {

				TextColor = W3CColorValue( arg[i] + 4, TRUE );

				if ( TextColor == W3C_COLOR_UNDEFINED ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Unknown color name: \"%s\"\r\n", arg[i] + 4 );
					goto cleanup_and_exit;
				}
			}
			else if ( !_wcsnicmp( arg[i], L"/XS", 3 ) && IsColonOrEquals( arg[i][3] ) ) {

				if ( ParseInt( arg[i] + 4, &TextSize, NULL ) ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Bad text size value: \"%s\"\r\n", arg[i] + 4 );
					goto cleanup_and_exit;
				}

				if ( ( TextSize < FONT_SIZE_MIN ) || ( TextSize > FONT_SIZE_MAX ) ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Text size must be %u to %u: %u\r\n", FONT_SIZE_MIN, FONT_SIZE_MAX, TextSize );
					goto cleanup_and_exit;
				}

				if ( TextY < 0 )
					TextY = 0 - ( ( TextSize * 3 ) / 2 );
			}
			else if ( !_wcsnicmp( arg[i], L"/XX", 3 ) && IsColonOrEquals( arg[i][3] ) ) {

				if ( ParseSignedInt( arg[i] + 4, &TextX, NULL ) ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Bad text X value: \"%s\"\r\n", arg[i] + 4 );
					goto cleanup_and_exit;
				}
			}
			else if ( !_wcsnicmp( arg[i], L"/XY", 3 ) && IsColonOrEquals( arg[i][3] ) ) {

				if ( ParseSignedInt( arg[i] + 4, &TextY, NULL ) ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Bad text Y value: \"%s\"\r\n", arg[i] + 4 );
					goto cleanup_and_exit;
				}
			}
			else if ( !_wcsnicmp( arg[i], L"/XF", 3 ) && IsColonOrEquals( arg[i][3] ) ) {

				if ( arg[i][4] == '\0' ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Missing typeface name\r\n" );
					goto cleanup_and_exit;
				}

				if ( wcslen( arg[i] + 4 ) > ( SHORT_BUF_LEN - 1 ) ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Typeface name too long: \"%s\"\r\n", arg[i] + 4 );
					goto cleanup_and_exit;
				}

				wcscpy_s( TextFace, SHORT_BUF_LEN, arg[i] + 4 );
			}
			else if ( !_wcsicmp( arg[i], L"/XI" ) )
				TextStyle |= FontStyleItalic;
			else if ( !_wcsicmp( arg[i], L"/XB" ) )
				TextStyle |= FontStyleBold;
			else if ( !_wcsicmp( arg[i], L"/XBI" ) )
				TextStyle |= FontStyleBoldItalic;
			else if ( !_wcsicmp( arg[i], L"/XU" ) )
				TextStyle |= FontStyleUnderline;
			else if ( !_wcsicmp( arg[i], L"/XN" ) )
				TextStyle = 0;
			else if ( !_wcsnicmp( arg[i], L"/X", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				wchar_t DefaultText[SHORT_BUF_LEN]		= L" \u2022 ScreenShot    %_isodate %_time";
				LPTSTR UserText			= arg[i] + j;

				if ( UserText[0] == '\0' ) {
					ExpandVariables( DefaultText, 1 );
					UserText = DefaultText;
				}
				
				size_t TextLen			= wcslen( UserText );
				size_t FaceNameLen		= wcslen( TextFace );

				_DrawTextArgs *NewArgs	= (_DrawTextArgs*) malloc( sizeof( _DrawTextArgs ) + ( TextLen + FaceNameLen + 2 ) * sizeof( wchar_t ) );
				_DrawTextArgs *This		= NULL;


				if ( NewArgs ) {

					wcscpy_s( &( NewArgs->Text ), TextLen + FaceNameLen + 2, UserText );
					NewArgs->Next		= NULL;
					NewArgs->X			= TextX;
					NewArgs->Y			= TextY;
					NewArgs->FaceName	= &( NewArgs->Text ) + TextLen + 1;
					wcscpy_s( NewArgs->FaceName, FaceNameLen + 1, TextFace );
					NewArgs->TextSize	= TextSize;
					NewArgs->TextStyle	= TextStyle;
					NewArgs->Color		= TextColor;
					NewArgs->Flags		= 0;

					if ( FirstTextArgs ) {
						This = FirstTextArgs;
						while (This->Next )
							This = This->Next;

						This->Next = NewArgs;
					}
					else
						FirstTextArgs = NewArgs;

					TextX = 0;
					TextY = TextY + ( TextSize * 3 ) / 2;
				}
			}
			else if ( !_wcsnicmp( arg[i], L"/Z", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( IsGraphicsExtension( arg[i] + j ) )
					wcscpy_s( DefaultExtension, EXT_BUF_LEN, arg[i] + j );
				else {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Unsupported extension: \"%s\"\r\n", arg[i] );
					goto cleanup_and_exit;
				}
			}
			else if ( arg[i][0] == '/' ) {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				goto cleanup_and_exit;
			}
			else if ( Filename[0] == '\0' ) {
				if ( _wcsicmp( arg[i], ClipName ) == 0 )
					wcscpy_s( Filename, MAX_FILENAME_LEN, ClipName );
				else if ( IsPrinter( arg[i] ) ) {
					wcscpy_s( Filename, MAX_FILENAME_LEN, arg[i] );
					for ( size_t i = 0; i < wcslen( PrintName ); i++ )
						Filename[i] = towupper( Filename[i] );
				}
				else
					QueryTrueName( arg[i], Filename );
			}
			else {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
				goto cleanup_and_exit;
			}

		}
	}


	rv	= 2;


	if ( DefaultExtension[0] == '\0' )
		GetDefaultExtension( DefaultExtension );

	
	if ( Filename[0] == '\0' )
		GenerateDefaultFilename( Filename, NULL, DefaultExtension );
	else if ( QueryIsDirectory( Filename ) )
		GenerateDefaultFilename( Filename, Filename, DefaultExtension );



	if ( Debug ) {
		Printf( L"   * Mode         : %d\r\n", Mode );
		Printf( L"   * Window title : '%s'\r\n", WindowTitle ? WindowTitle : NullString );
		Printf( L"   * Window class : '%s'\r\n", WindowClass ? WindowClass : NullString );
		Printf( L"   * PID          : 0x%04X\r\n", Pid );
		Printf( L"   * Filename     : '%s'\r\n", Filename );
	}


	if ( Mode == MODE_DESKTOP ) {
		//		get monitor coordinates:

		Left	= 0;
		Top		= 0;
		Width	= GetSystemMetrics( SM_CXSCREEN );
		Height	= GetSystemMetrics( SM_CYSCREEN );

		_GetMonitorCoordsStruct Monitors;
		GetMonitorCoords( &Monitors );


		if ( ( Monitor >= Monitors.NumMonitors ) && ( Monitor < ALL_MONITORS ) )
			Monitor = Monitors.Default;

		if ( Monitor < Monitors.NumMonitors ) {
			Left	= Monitors.Left[Monitor];
			Top		= Monitors.Top[Monitor];
			Width	= Monitors.Width[Monitor];
			Height	= Monitors.Height[Monitor];
		}
		else {
			Left	= GetSystemMetrics( SM_XVIRTUALSCREEN );
			Top		= GetSystemMetrics( SM_YVIRTUALSCREEN );
			Width	= GetSystemMetrics( SM_CXVIRTUALSCREEN );
			Height	= GetSystemMetrics( SM_CYVIRTUALSCREEN );
		}
	}
	else {
		//		find window:

		switch ( Mode ) {
		case MODE_FG_WINDOW :
			WindowHandle = GetForegroundWindow();

			if ( !WindowHandle ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Can\u2019t get foreground window!\r\n" );
				goto cleanup_and_exit;
			}
				
			break;

		case MODE_HWND :

			if ( !IsWindow( WindowHandle ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Not a valid window handle: 0x%08X\r\n", WindowHandle );
				goto cleanup_and_exit;
			}

			if ( !IsWindowVisible( WindowHandle ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Window not visible: 0x%08X\r\n", WindowHandle );
				goto cleanup_and_exit;
			}

			break;

		default :
			WindowHandle = MyFindWindow( Pid, WindowTitle, WindowClass );

			if ( !WindowHandle ) {
				DisplayWindowNotFoundError( Pid, WindowTitle, WindowClass );
				goto cleanup_and_exit;
			}
		}


		if ( WindowHandle )
			if ( IsHWAWindow( WindowHandle ) )
				ClientAreaOnly = WindowsOwnContext = FALSE;


		Minimized			= IsIconic( WindowHandle );
		ForegroundWindow	= GetForegroundWindow();
		ForceForegroundWindow( WindowHandle );
		Sleep( Minimized ? 150 : 75 );


		memset( &WindowInfo, 0x00, sizeof( WINDOWINFO ) );
		WindowInfo.cbSize = sizeof( WINDOWINFO );
		GetWindowInfo( WindowHandle, &WindowInfo );

		if ( WindowsOwnContext ) {
			Left	= 0;
			Top		= 0;
			Width	= WindowInfo.rcWindow.right  - WindowInfo.rcWindow.left + 1;
			Height	= WindowInfo.rcWindow.bottom - WindowInfo.rcWindow.top  + 1;
		}
		else if ( ClientAreaOnly ) {
			Left	= 0;
			Top		= 0;
			Width	= WindowInfo.rcClient.right  - WindowInfo.rcClient.left + 1;
			Height	= WindowInfo.rcClient.bottom - WindowInfo.rcClient.top  + 1;
		}
		else {
			Left	= WindowInfo.rcClient.left;
			Top		= WindowInfo.rcClient.top;
			Width	= WindowInfo.rcClient.right  - WindowInfo.rcClient.left + 1;
			Height	= WindowInfo.rcClient.bottom - WindowInfo.rcClient.top  + 1;
		}
	}


	if ( IsPrinter( Filename ) ) {			//	if we're printing the image, don't resize it; ignore any /R
		ResizeArgs.NewWidth		= 0;
		ResizeArgs.NewHeight	= 0;
		ResizeArgs.Percent		= FALSE;
	}


	if ( Debug ) {
		Printf( L"   * Window handle = 0x%08X\r\n", WindowHandle );
		Printf( L"   * Left = %d \tWidth = %d\r\n", Left, Width );
		Printf( L"   * Top = %d \tHeight = %d\r\n", Top, Height );
		Printf( L"   * NewWidth = %u   NewHeight = %u   Percent = %u\r\n", ResizeArgs.NewWidth, ResizeArgs.NewHeight, ResizeArgs.Percent );
	}


	if ( WindowHandle && WindowsOwnContext )
		WindowDC = GetWindowDC( WindowHandle );
	else if ( WindowHandle && ClientAreaOnly )
		WindowDC = GetDC( WindowHandle );
	else
		WindowDC = GetDC( NULL );


	if ( WindowDC ) {

		rv = SaveWindowPic( WindowDC, Filename, Overwrite, Left, Top, Width, Height, &ResizeArgs, FirstTextArgs );

		if ( rv ) {
			GetGdiplusErrorText( rv, ErrorText );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error saving to \"%s\":\r\n", Filename );
			Qprintf( ERRHANDLE, L"    %s\r\n", ErrorText );
			wcscpy_s( LastFilename, MAX_FILENAME_LEN, ErrorString );
			rv = 2;
		}
		else {
			if ( !Quiet )
				Printf( L"Saved screen shot to \"%s\".\r\n", Filename );

			wcscpy_s( LastFilename, MAX_FILENAME_LEN, Filename );
			rv = 0;
		}

		ReleaseDC( WindowHandle, WindowDC );
	}
	else {
		Qprintf( ERRHANDLE, L"Can\u2019t get a DC for window 0x%08X\r\n", WindowHandle );
		rv = 2;
	}
	
	
	if ( ForegroundWindow )
		ForceForegroundWindow( ForegroundWindow );

	/*
	if ( Minimized )
		ShowWindow( WindowHandle, SW_SHOWMINIMIZED );
	*/


cleanup_and_exit:

	if ( FirstTextArgs ) {
		_DrawTextArgs *This		= FirstTextArgs;
		_DrawTextArgs *Next		= NULL;

		while ( This ) {
			Next = This->Next;
			free( This );
			This = Next;
		}
	}

	return rv;
}



DLLExports int WINAPI _lastssfile( LPTSTR outString )
{
	if ( !outString )
		return 1;


	wcscpy_s( outString, MAX_FILENAME_LEN, LastFilename );
	return 0;
}



#ifdef			SSWINLIST

int RepeatChar( wchar_t ch, unsigned int n )
{
	wchar_t Str[1025] = L"";


	if ( n > 1024 )
		n = 1024;

	for ( unsigned int i = 0; i < n; i++ )
		Str[i] = ch;

	Str[n] = '\0';

	QPuts( Str );
	return 0;
}



int ShowSpc( unsigned int n )
{
	return RepeatChar( 0x20, n );
}



int ShowStringWidth( LPTSTR String, unsigned int Width, unsigned int Margin )
{
	if ( !String )
		return -1;


	if ( wcslen( String ) >= (size_t) Width )
		Printf( L"%.*s", Width, String );
	else {
		QPuts( String );
		ShowSpc( Width - (unsigned int) wcslen( String ) );
	}

	if ( Margin )
		ShowSpc( Margin );

	return 0;
}



struct _WinListArgs {
	LPTSTR TitleSpec;
	LPTSTR ClassSpec;
	DWORD Pid;
	BOOL IncludeHidden;
	BOOL LineNumbers;
	unsigned int TitleWidth;
	unsigned int ClassWidth;
	unsigned int Count;
};



BOOL CALLBACK _WinListProc( HWND hWnd, LPARAM Args )
{
	if ( !Args )
		return FALSE;


	_WinListArgs *args = (_WinListArgs*) Args;

	wchar_t Title[SHORT_BUF_LEN]		= L"";
	wchar_t Class[SHORT_BUF_LEN]		= L"";
	DWORD Pid							= 0;


	if ( args->IncludeHidden == FALSE )
		if ( !IsWindowVisible( hWnd ) )
			return TRUE;


	GetWindowText( hWnd, Title, SHORT_BUF_LEN );
	RealGetWindowClass( hWnd, Class, SHORT_BUF_LEN );
	GetWindowThreadProcessId( hWnd, &Pid );


	if ( args->TitleSpec )
		if ( args->TitleSpec[0] )
			if ( WildcardComparison( args->TitleSpec, Title, 0, 1 ) )
				return TRUE;

	if ( args->ClassSpec )
		if ( args->ClassSpec[0] )
			if ( WildcardComparison( args->ClassSpec, Class, 0, 1 ) )
				return TRUE;

	if ( args->Pid )
		if ( Pid != args->Pid )
			return TRUE;


	if ( args->LineNumbers )
		Printf( L"%4u  ", args->Count + 1 );

	Printf( L"%08X  %04X  ", hWnd, Pid );
	ShowStringWidth( Title, args->TitleWidth, 2 );
	ShowStringWidth( Class, args->ClassWidth, 0 );
	CrLf();

	args->Count += 1;
	return TRUE;
}



int WinListHeader( _WinListArgs *Args )
{
	if ( !Args )
		return -1;


	CrLf();

	if ( Args->LineNumbers )
		QPuts( L"   #  " );

	QPuts( L"HWND      PID   " );
	ShowStringWidth( L"Title", Args->TitleWidth, 2 );
	ShowStringWidth( L"Class", Args->ClassWidth, 0 );

	CrLf();

	if ( Args->LineNumbers )
		QPuts( L"----  " );

	QPuts( L"--------  ----  " );
	RepeatChar( '-', Args->TitleWidth );
	ShowSpc( 2 );
	RepeatChar( '-', Args->ClassWidth );

	CrLf();

	return 0;
}



int WinListFooter( _WinListArgs *Args )
{
	if ( !Args )
		return -1;


	if ( Args->Count == 0 )
		QPuts( L"\r\nNo windows found.\r\n" );
	else if ( Args->Count == 1 )
		Printf( L"\r\n%u window found.\r\n", Args->Count );
	else
		Printf( L"\r\n%u windows found.\r\n", Args->Count );

	return 0;
}



int FigureWinListColumnWidths( _WinListArgs *Args, unsigned int Columns )
{
	if ( !Args )
		return -1;


	Columns = Columns - ( 10 + 6 + 2 );			// hwnd + pid + spaces;
	if ( Args->LineNumbers )
		Columns -= 6;

	Args->TitleWidth = ( Columns * 4 ) / 7;
	Args->ClassWidth = Columns - Args->TitleWidth;

	return 0;
}



DLLExports int WINAPI sswinlist( LPTSTR inString )
{
	if ( !inString )
		return ERROR_INVALID_PARAMETER;


	_WinListArgs Args;

	LPTSTR TitleSpec		= NULL;
	LPTSTR ClassSpec		= NULL;
	unsigned int Pid		= 0;

	BOOL LineNumbers		= FALSE;
	BOOL IncludeHidden		= FALSE;
	BOOL Quiet				= FALSE;

	unsigned int Width		= 0;



	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {

			if ( !wcsncmp( arg[i], L"/?", 2 ) )
				return ShowCmdHelp( SSWinListHelpText, ( arg[i][2] == '?' ) );
			else if ( !_wcsnicmp( arg[i], L"/C", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &Width, NULL ) ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad columns value: \"%s\"\r\n", arg[i] );
					return 1;
				}

				if ( Width && ( ( Width < MIN_COLUMNS ) || ( Width > MAX_COLUMNS ) ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Columns must be %u - %u: %u\r\n", MIN_COLUMNS, MAX_COLUMNS, Width );
					return 1;
				}
			}
			else if ( !_wcsicmp( arg[i], L"/H" ) )
				IncludeHidden = TRUE;
			else if ( !_wcsnicmp( arg[i], L"/K", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] == '\0' )
					ClassSpec = NULL;
				else
					ClassSpec = arg[i] + j;
			}
			else if ( !_wcsicmp( arg[i], L"/L" ) )
				LineNumbers = TRUE;
			else if ( !_wcsnicmp( arg[i], L"/P", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &Pid, NULL ) ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad PID value: \"%s\"\r\n", arg[i] );
					return 1;
				}
			}
			else if ( !_wcsicmp( arg[i], L"/Q" ) )
				Quiet = TRUE;
			else if ( arg[i][0] == '/' ) {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}
			else if ( !TitleSpec )
				TitleSpec = arg[i];
			else {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
				return 1;
			}

		}
	}


	Args.TitleSpec		= TitleSpec;
	Args.ClassSpec		= ClassSpec;
	Args.Pid			= Pid;
	Args.IncludeHidden	= IncludeHidden;
	Args.LineNumbers	= LineNumbers;
	Args.TitleWidth		= 10;
	Args.ClassWidth		= 10;
	Args.Count			= 0;



	if ( !Width ) {
		Width = DEFAULT_COLUMNS;

		if ( ConsoleHandle( STD_OUTPUT_HANDLE ) ) {
			Width = ConsoleWidth() - 1;

			if ( ( Width < MIN_COLUMNS ) || ( Width > MAX_COLUMNS ) )
				Width = DEFAULT_COLUMNS;
		}
	}


	FigureWinListColumnWidths( &Args, Width );


	if ( !Quiet )
		WinListHeader( &Args );


	EnumWindows( _WinListProc, (LPARAM) &Args );


	if ( !Quiet )
		WinListFooter( &Args );


	if ( Args.Count )
		return 0;
	else
		return 2;
}

#endif		//	SSWINLIST



LPTSTR UsageText( const wchar_t *inString )
{
	if ( !inString )
		return NULL;


	wchar_t *ch = wcschr( (wchar_t*) inString, '\n' );


	if ( !ch )
		return (LPTSTR) inString;

	while ( *ch == '\n' )
		ch++;

	return ch;
}



DLLExports int WINAPI Usage( LPTSTR inString, LPTSTR outString )
{
	if ( ( inString[0] == '@' ) || ( inString[0] == '_' ) )
		return 0;

	for ( int i = 0; i < NumHelpTexts; i++ ) {
		if ( !_wcsicmp( inString, HelpFeature[i] ) ) {
			wcscpy_s( outString, SHORT_BUF_LEN, UsageText( HelpText[i] ) );
			return 1;
		}
	}

	return 0;
}



#ifndef NEWHELP_VERSION

//		Simpler MessageBox() help, useful for smaller plugins which don't merit a full HTML help system:

DLLExports int WINAPI Help( LPTSTR inString )
{
	wchar_t WindowTitle[SHORT_BUF_LEN]		= L"";


	wsprintf( WindowTitle, L"%s Plugin Help", PLUGIN_NAME );

	for ( int i = 0; i < NumHelpTexts; i++ ) {
		if ( !_wcsicmp( inString, HelpFeature[i] ) ) {
			MessageBox( GetForegroundWindow(), HelpText[i], WindowTitle, MB_OK );
			return 1;
		}
	}

	return 0;
}


#endif		//	NEWHELP_VERSION


/*

		1.1.0.0		2021-11-11
					Rewritten to grab the entire desktop, not just one window, in order to
					capture programs like Chrome and Edge which use video acceleration.
					/E reverts to the old behavior.

		1.1.1.0		2021-12-21
					Fixed an issue calling TakeCommandIPC() from 64-bit TCC.

		1.1.2.0		2021-12-22
					No longer call TakeCommandIPC() dynamically.  That hasn't been
					necessary for a long, long time.

		1.1.2.2		2023-10-12
					Updated the plugin's web address.

		1.1.2.3		2023-10-15
					Tweaked ShowCmdHelp() to include VER_PATCH.

		1.1.3.0		2023-12-06
					In the default mode (sans /C and /E), does a better job of finding
					the window's edges.
					Replaced GetTakeCommandWindow() with FindMyWindow.cpp; now /T should
					find Windows Terminal, ConEmu, and Console2 windows.
					If the window is of a class known to use hardware acceleration
					(Google Chrome, MS Edge, Firefox, Windows Terminal), /C and /E will
					be disabled automatically.
					Cosmetic:  Changed a bunch of ASCII apostrophes to fancy Unicode quotes,
					mostly in error messages.

		1.1.3.1		2024-03-13
					Minor tweaks to ForceForegroundWindow().

*/
